"""
Generally the SciTSR format consists of three different files.
chunks/xyz.chunk - These simply lists the cell bounding boxes with their corresponding text
structure/xyz.structure - These are structure files. They should translate the cell bboxes to a table structure.
img/xyz.png - the image files
"""
from typing import Tuple, List

from converter.elements import ConversionStrategy, ConverterDocument, SciTSRConverterDocument
from docrecjson.elements import Document, PolygonRegion, Cell

from loguru import logger


class SciTsrStrategy(ConversionStrategy):
    def initialize(self, original: ConverterDocument) -> ConverterDocument:
        # todo determine the image resolution from the file instead of using
        document: Document = Document.empty(original.filename, original_image_size=(1241, 1754))

        original.shared_file_format_document = document
        return original

    def add_metadata(self, original: ConverterDocument) -> ConverterDocument:
        document: Document = original.shared_file_format_document
        document.add_metadata({"creator": "shared-file-converter"})
        document.add_metadata({"creator": "sciTSR"})
        return original

    # todo this method was more intended as the polygonRegion adder method and not table + cell region adder
    # todo add method which handles table structure options separately
    # change ConversionStrategy for this
    def add_regions(self, original: SciTSRConverterDocument) -> ConverterDocument:
        document: Document = original.shared_file_format_document

        """
        The chunks and Structure dictionaries do not match 100%, since the chunks dict are more of a description of text
        blocks. 
        We match the chunks and structure entries by position and assume that there must be a chunk for each structure. 
        This is necessary because the structure does not contain any positional information.
        Since we assume that the structure files describe the table cells, the remaining chunks are text blocks on the 
        page which are not in the table. Therefore, they are not added as cells, but as ordinary text regions.
        """
        chunks_dct: dict = original.tmp_type
        structure_dct: dict = original.structure_dct

        if len(structure_dct) != 1:
            logger.error(
                "Unexpected length of SciTSR structure file. "
                "Please review weather all elements are processed correctly.")
            raise RuntimeError("Wrong length of SciTSR Structure annotation.")

        unified_chunk_and_structure: dict = {"cells": []}
        # these are all text blocks which are not part of the table, but are still in the SciTSR image and pdf files
        # these are mostly the table descriptions
        chunks_which_are_not_cells: dict = {"chunks": []}
        chunks: list = chunks_dct['chunks']
        structures_complete: list = structure_dct['cells']
        non_empty_structures: list = []
        # these are the structure entries without any kind of text.
        # these structure entries do not have a chunk representation with coordinates
        # todo determine the coordinates of the structure somehow
        empty_structures: list = []

        for i in range(0, len(structures_complete)):
            structure_element_with_index: dict = next(x for x in structures_complete if x["id"] == i)
            if not structure_element_with_index["content"]:
                empty_structures.append(structure_element_with_index)
            else:
                non_empty_structures.append(structure_element_with_index)

        for i, chunks_element in enumerate(chunks):
            if i >= len(non_empty_structures):
                chunks_which_are_not_cells["chunks"].append(chunks_element)
            else:
                structure_element: dict = non_empty_structures[i]
                chunks_element["structure_tex"] = structure_element["tex"]
                chunks_element["structure_content"] = structure_element["content"]
                chunks_element["start_row"] = structure_element["start_row"]
                chunks_element["end_row"] = structure_element["end_row"]
                chunks_element["start_col"] = structure_element["start_col"]
                chunks_element["end_col"] = structure_element["end_col"]

                unified_chunk_and_structure["cells"].append(chunks_element)

        # todo this is the issue with empty cells -> no chunks representation
        # if len(unified_chunk_and_structure["cells"]) != len(structures):
        #     raise RuntimeError("Expected all structure elements to be Part of a chunk. Mismatching array length.")

        cells: List[Cell] = []
        for element in unified_chunk_and_structure["cells"]:
            pos: list = element['pos']
            point_1 = self._convert_pts_to_pxl_coordinates((pos[0], pos[2]))
            point_2 = self._convert_pts_to_pxl_coordinates((pos[1], pos[3]))

            cell: Cell = document.add_cell(self._span_polygon(point_1, point_2), element["start_row"],
                                           element["end_row"], element["start_col"], element["end_col"],
                                           element["text"], source="ground_truth")
            cells.append(cell)

        document.add_table(self.get_table_coordinates_from_cells(cells), cells, source="ground_truth")

        document.set_source_for_adding("ground_truth")
        for element in chunks_which_are_not_cells["chunks"]:
            pos: list = element['pos']
            txt: str = element['text']

            point_1 = self._convert_pts_to_pxl_coordinates((pos[0], pos[2]))
            point_2 = self._convert_pts_to_pxl_coordinates((pos[1], pos[3]))

            region: PolygonRegion = document.add_region(self._span_polygon(point_1, point_2), "table description")
            document.add_text(txt, region)

        return original

    def get_table_coordinates_from_cells(self, cells: List[Cell]) -> list:
        """
        Computes the cell bounding box based on the already extracted cells.
        It just needs to compute the lower left coordinate, as well as the upper right coordinate.
        The remaining coordinates can be computed with _span_polygon
        :param cells: all cells in the tables. The cell's bounding box can be accessed via cell.bounding_box.
                      The single coordinates are in the order as they are returned by _span_polygon.
                      This is because _span_polygon was already used for the cell bounding box creation.
        :return: all four rectangle coordinates of the table bounding box
        """
        all_x_values = []
        all_y_values = []

        for cell in cells:
            for point in cell.bounding_box.polygon:
                all_x_values.append(point[0])
                all_y_values.append(point[1])

        # lower left coordinate = min x coordinate + max y coordinate
        # upper right coordinate = max x coordinate + min y coordinate
        max_x = max(all_x_values)
        min_x = min(all_x_values)
        max_y = max(all_y_values)
        min_y = min(all_y_values)

        return self._span_polygon((min_x, max_y), (max_x, min_y))

    @staticmethod
    def _span_polygon(point1: Tuple, point2: Tuple) -> list:
        """
        The sci tsr polygon bounding boxes do not have the necessary coordinate structure for the shared-file-format.
        The coordinates are simply the lower left of the bbox and the upper right of the bbox.
        But this is sufficient to construct the right square coordinates.
        It's important that the coordinates are in the right order because the shared-file-format assumes that the last
        coordinates are connected.
        :param point1:  lower left coordinate
        :param point2: upper right coordinate
        :return: a list of four coordinates. with index:
            0 = lower left
            1 = lower right
            2 = upper right
            3 = upper left
        """
        # written not in a single statement for readability
        # noinspection PyListCreation
        polygon_points: list = []

        polygon_points.append(point1)
        # point2.x, point1.y = lower right
        polygon_points.append((point2[0], point1[1]))
        polygon_points.append(point2)
        # point1.x, point2.y = upper left
        polygon_points.append((point1[0], point2[1]))

        return polygon_points

    @staticmethod
    def _convert_pts_to_pxl_coordinates(coord: Tuple):
        """
        Theoretically, it would be possible to make this conversion more flexible by dynamically reading the
        sizes of the images.
        But based on the examples I saw, i assume that each picture hase a fixed size which simplifies development
        drastically. To make dynamic conversoin implementable afterwards, I write out the concrete numbers.
        :return:
        """
        # default A4 values for pdf
        pts_max_width = 595.276
        pts_max_height = 841.89

        pxl_max_width = 1241
        pxl_max_height = 1754

        width_multiplicator = pxl_max_width / pts_max_width
        heigh_multiplicator = pxl_max_height / pts_max_height

        # these are the coordinates from (0, 0) at our bottom left, but we want the (0, 0) at the top left
        bottom_left_x = coord[0] * width_multiplicator
        bottom_left_y = coord[1] * heigh_multiplicator

        return int(bottom_left_x), int(pxl_max_height - bottom_left_y)
